from typing import Optional


class NoSuchProcess(BaseException): ...
class TimeoutExpired(BaseException): ...

STATUS_ZOMBIE: int

class Process:

    def __init__(self, pid: int): ...
    def is_running(self) -> bool: ...
    def status(self) -> int: ...
    def wait(self, timeout: Optional[int] = None) -> None: ...
    def kill(self) -> None: ...
