from typing import Any, Iterator


class RunningCommand:
    def __init__(self, *args: Any, **kwargs: Any): ...

    def __iter__(self) -> Iterator[str]: ...

class Command:
    def __init__(self, *args: Any, **kwargs: Any): ...

    def bake(self, *args: Any, **kwargs: Any) -> "Command": ...

    def __call__(self, *args: Any, **kwargs: Any) -> RunningCommand: ...

class SignalException_SIGTERM(BaseException): ...
